from six.moves import builtins

__all__ = (
    'display',
)


def display(obj, **kwargs):
    # Perform lazy-import for fast initialization.
    # Note: pd/np are likely to be already imported by user codes!
    import pandas as pd
    import numpy as np
    if isinstance(obj, pd.DataFrame):
        r = (b'html', obj.to_html().encode('utf8'))
        builtins._sorna_emit(r)
        return
    if isinstance(obj, np.ndarray):
        df = pd.DataFrame(obj, **kwargs)
        r = (b'html', df.to_html().encode('utf8'))
        builtins._sorna_emit(r)
        return
    raise TypeError('Unsupported object type for display.')



import asyncio
import functools
import json
import logging
import os
from pathlib import Path
import shutil
import site
import tempfile

import janus
from jupyter_client import KernelManager
from jupyter_client.kernelspec import KernelSpecManager

from .. import BaseRunner

log = logging.getLogger()

DEFAULT_PYFLAGS = []


class Runner(BaseRunner):

    log_prefix = 'python-kernel'
    default_runtime_path = '/usr/bin/python'
    default_child_env = {
        'TERM': 'xterm',
        'LANG': 'C.UTF-8',
        'SHELL': '/bin/ash' if Path('/bin/ash').is_file() else '/bin/bash',
        'USER': 'work',
        'HOME': '/home/work',
        'PATH': ':'.join([
            '/usr/local/nvidia/bin',
            '/usr/local/cuda/bin',
            '/usr/local/sbin',
            '/usr/local/bin',
            '/usr/sbin',
            '/usr/bin',
            '/sbin',
            '/bin',
        ]),
        'LD_LIBRARY_PATH': os.environ.get('LD_LIBRARY_PATH', ''),
        'LD_PRELOAD': os.environ.get('LD_PRELOAD', ''),
        'PYTHONPATH': site.USER_SITE,
    }

    def __init__(self):
        super().__init__()
        self.sentinel = object()
        self.input_queue = None
        self.output_queue = None

        # Add sitecustomize.py to site-packages directory.
        # No permission to access global site packages, we use user local directory.
        input_src = Path(os.path.dirname(__file__)) / 'sitecustomize.py'
        # pkgdir = Path(site.getsitepackages()[0])
        pkgdir = Path(site.USER_SITE)
        pkgdir.mkdir(parents=True, exist_ok=True)
        shutil.copy(str(input_src), str(pkgdir / 'sitecustomize.py'))

        # Detect ipython kernel spec for backend.ai and start it.
        kernelspec_mgr = KernelSpecManager()
        kspecs = kernelspec_mgr.get_all_specs()
        for kname in kspecs:
            if kname.startswith('backendai-python'):
                log.info('starting ipykernel...')
                self.kernel_mgr = KernelManager(kernel_name=kname)
                self.kernel_mgr.start_kernel()
                if not self.kernel_mgr.is_alive():
                    log.error('query mode is disabled: '
                              'failed to start jupyter kernel')
                else:
                    self.kernel_client = self.kernel_mgr.client()
                    self.kernel_client.start_channels(shell=True, iopub=True,
                                                      stdin=True, hb=True)
                    try:
                        self.kernel_client.wait_for_ready(timeout=5)
                    except RuntimeError:
                        # Clean up for client and kernel will be done in `shutdown`.
                        log.error('jupyter channel is not active!')
                        self.kernel_mgr = None
                break
        else:
            log.info('query mode is disabled: '
                     'no jupyter kernelspec found')
            self.kernel_mgr = None

    async def init_with_loop(self):
        self.input_queue = janus.Queue(loop=self.loop)
        self.output_queue = janus.Queue(loop=self.loop)

        # We have interactive input functionality!
        self._user_input_queue = janus.Queue(loop=self.loop)
        self.user_input_queue = self._user_input_queue.async_q

    async def shutdown(self):
        if self.kernel_mgr and self.kernel_mgr.is_alive():
            log.info('shutting down ipykernel...')
            self.kernel_client.stop_channels()
            self.kernel_mgr.shutdown_kernel()
            assert not self.kernel_mgr.is_alive(), 'ipykernel failed to shutdown'

    async def build_heuristic(self) -> int:
        if Path('setup.py').is_file():
            cmd = [
                self.runtime_path, *DEFAULT_PYFLAGS,
                '-m', 'pip', 'install', '-e', '.',
            ]
            return await self.run_subproc(cmd)
        else:
            log.warning('skipping the build phase due to missing "setup.py" file')
            return 0

    async def execute_heuristic(self) -> int:
        if Path('main.py').is_file():
            cmd = [
                self.runtime_path, *DEFAULT_PYFLAGS,
                'main.py',
            ]
            return await self.run_subproc(cmd)
        else:
            log.error('cannot find the main script ("main.py").')
            return 127

    async def query(self, code_text) -> int:
        if self.kernel_mgr is None:
            log.error('query mode is disabled: '
                      'failed to start jupyter kernel')
            return 127

        log.debug('executing in query mode...')
        loop = asyncio.get_event_loop()

        def output_hook(msg):
            if msg['msg_type'] == 'stream':
                content = msg['content']
                loop.call_soon_threadsafe(self.outsock.send_multipart,
                                          [content['name'].encode('ascii'),
                                           content['text'].encode('utf-8')])
            elif msg['msg_type'] == 'error':
                content = msg['content']
                loop.call_soon_threadsafe(self.outsock.send_multipart,
                                          [b'stderr',
                                           content['traceback'][0].encode('utf-8')])

        def stdin_hook(msg):
            if msg['msg_type'] == 'input_request':
                prompt = msg['content']['prompt']
                password = msg['content']['password']
                if prompt:
                    loop.call_soon_threadsafe(self.outsock.send_multipart,
                                              [b'stdout', prompt.encode('utf-8')])
                loop.call_soon_threadsafe(
                    self.outsock.send_multipart,
                    [b'waiting-input',
                     json.dumps({'is_password': password}).encode('utf-8')])
                user_input = self._user_input_queue.sync_q.get()
                self.kernel_client.input(user_input)

        # Run jupyter kernel's blocking execution method in an executor pool.
        await loop.run_in_executor(
            None,
            functools.partial(self.kernel_client.execute_interactive,
                              code_text, allow_stdin=True, timeout=2,
                              output_hook=output_hook, stdin_hook=stdin_hook)
        )
        return 0

    async def complete(self, data):
        # TODO: implement with jupyter_client
        '''
        matches = []
        self.outsock.send_multipart([
            b'completion',
            json.dumps(matches).encode('utf8'),
        ])
        '''
        # self.kernel_mgr.complete(data, len(data))

    async def interrupt(self):
        # TODO: implement with jupyter_client
        self.kernel_mgr.interrupt_kernel()

    async def start_service(self, service_info):
        if service_info['name'] == 'jupyter':
            with tempfile.NamedTemporaryFile(
                    'w', encoding='utf-8', suffix='.py', delete=False) as config:
                print('c.NotebookApp.allow_root = True', file=config)
                print('c.NotebookApp.ip = "0.0.0.0"', file=config)
                print('c.NotebookApp.port = {}'.format(service_info['port']),
                      file=config)
                print('c.NotebookApp.token = ""', file=config)
            return [
                self.runtime_path, '-m', 'jupyter', 'notebook',
                '--no-browser',
                '--config', config.name,
            ], {}
        elif service_info['name'] == 'ipython':
            return [
                self.runtime_path, '-m', 'IPython',
            ], {}
        elif service_info['name'] == 'digits':
            return [
                self.runtime_path, '-m', 'digits',
            ], {}
        elif service_info['name'] == 'tensorboard':
            return [
                self.runtime_path, '-m', 'tensorboard.main',
            ], {}



import builtins as builtin_mod
import code
from functools import partial
from io import IOBase, UnsupportedOperation
import json
import logging
import sys
import traceback
import threading
import types

from IPython.core.completer import Completer

import getpass

from .display import display
from .types import (
    ConsoleRecord, MediaRecord, HTMLRecord,
)


log = logging.getLogger()


class ConsoleOutput(IOBase):

    def __init__(self, emit, stream_type):
        self._emit = emit
        self._stream_type = stream_type

    def readable(self):
        return False

    def writable(self):
        return True

    def seekable(self):
        return False

    def fileno(self):
        raise OSError(f'{self._stream_type} has no file descriptor '
                      'because it is a virtual console.')

    def read(self, *args, **kwargs):
        raise UnsupportedOperation()

    def write(self, s):
        if self.closed:
            raise ValueError('Cannot write to the closed console.')
        if isinstance(s, str):
            s = s.encode('utf8')
        self._emit(ConsoleRecord(self._stream_type, s))

    def flush(self):
        pass

    def isatty(self):
        return True


class PythonInprocRunner(threading.Thread):
    '''
    A thin wrapper for REPL.

    It creates a dummy module that user codes run and keeps the references to
    user-created objects (e.g., variables and functions).
    '''

    def __init__(self, input_queue, output_queue, user_input_queue, sentinel):
        super().__init__(name='InprocRunner', daemon=True)

        # for interoperability with the main asyncio loop
        self.input_queue = input_queue
        self.output_queue = output_queue
        self.user_input_queue = user_input_queue
        self.sentinel = sentinel

        self.stdout = ConsoleOutput(self.emit, 'stdout')
        self.stderr = ConsoleOutput(self.emit, 'stderr')

        # Initialize user module and namespaces.
        user_module = types.ModuleType(
            '__main__',
            doc='Automatically created module for the interactive shell.')
        user_module.__dict__.setdefault('__builtin__', builtin_mod)
        user_module.__dict__.setdefault('__builtins__', builtin_mod)
        self.user_module = user_module
        self.user_ns = user_module.__dict__

        self.completer = Completer(namespace=self.user_ns, global_namespace={})
        self.completer.limit_to__all__ = True

    def run(self):
        # User code is executed in a separate thread.
        while True:
            code_text = self.input_queue.get()
            self.input_queue.task_done()

            # Set Backend.AI Media handler
            self.user_module.__builtins__._sorna_emit = self.emit
            self.user_module.__builtins__.display = display

            # Override interactive input functions
            self.user_module.__builtins__.input = self.handle_input
            getpass.getpass = partial(self.handle_input, password=True)

            try:
                code_obj = code.compile_command(code_text, symbol='exec')
            except (OverflowError, IndentationError, SyntaxError,
                    ValueError, TypeError, MemoryError):
                exc_type, exc_val, tb = sys.exc_info()
                user_tb = type(self).strip_traceback(tb)
                err_str = ''.join(traceback.format_exception(exc_type, exc_val,
                                                             user_tb))
                hdr_str = 'Traceback (most recent call last):\n' \
                        if not err_str.startswith('Traceback ') else ''
                self.stderr.write(hdr_str + err_str)
                self.output_queue.put(self.sentinel)
            else:
                sys.stdout, orig_stdout = self.stdout, sys.stdout
                sys.stderr, orig_stderr = self.stderr, sys.stderr
                try:
                    exec(code_obj, self.user_ns)
                except KeyboardInterrupt:
                    print('Interrupted!', file=self.stderr)
                except Exception:
                    # strip the first frame
                    exc_type, exc_val, tb = sys.exc_info()
                    user_tb = type(self).strip_traceback(tb)
                    traceback.print_exception(exc_type, exc_val, user_tb,
                                              file=sys.stderr)
                finally:
                    sys.stdout = orig_stdout
                    sys.stderr = orig_stderr
                    self.output_queue.put(self.sentinel)

    def handle_input(self, prompt=None, password=False):
        if prompt is None:
            prompt = 'Password: ' if password else ''
        # Use synchronous version of ZeroMQ sockets
        if prompt:
            self.output_queue.put([
                b'stdout',
                prompt.encode('utf8'),
            ])
        self.output_queue.put([
            b'waiting-input',
            json.dumps({'is_password': password}).encode('utf8'),
        ])
        data = self.user_input_queue.get()
        return data

    def complete(self, data):
        # This method is executed in the main thread.
        state = 0
        matches = []
        while True:
            ret = self.completer.complete(data['line'], state)
            if ret is None:
                break
            matches.append(ret)
            state += 1
        return matches

    def emit(self, record):
        if isinstance(record, ConsoleRecord):
            assert record.target in ('stdout', 'stderr')
            self.output_queue.put([
                record.target.encode('ascii'),
                record.data,
            ])
        elif isinstance(record, MediaRecord):
            self.output_queue.put([
                b'media',
                json.dumps({
                    'type': record.type,
                    'data': record.data,
                }).encode('utf8'),
            ])
        elif isinstance(record, HTMLRecord):
            self.output_queue.put([
                b'html',
                record.html.encode('utf8'),
            ])
        elif isinstance(record, tuple):  # raw output
            assert isinstance(record[0], bytes), 'emit: Raw record must be bytes.'
            assert isinstance(record[1], bytes), 'emit: Raw record must be bytes.'
            self.output_queue.put(record)
        else:
            raise TypeError('Unsupported record type.')

    @staticmethod
    def strip_traceback(tb):
        while tb is not None:
            frame_summary = traceback.extract_tb(tb, limit=1)[0]
            if frame_summary[0] == '<input>':
                break
            tb = tb.tb_next
        return tb



from collections import defaultdict
from decimal import Decimal, ROUND_DOWN
import enum
import io
import json
import logging
import operator
from pathlib import Path
import pkg_resources
import sys
from typing import Container, Collection, Mapping, Sequence

import attr
import psutil

from ai.backend.common.utils import readable_size_to_bytes
from .accelerator import AbstractAcceleratorInfo, ProcessorIdType, accelerator_types
from .vendor.linux import libnuma

log = logging.getLogger('ai.backend.agent.resources')


@attr.s(auto_attribs=True, slots=True)
class Share:
    device_id: ProcessorIdType
    share: Decimal


class MountPermission(enum.Enum):
    READ_ONLY = 'ro'
    READ_WRITE = 'rw'


@attr.s(auto_attribs=True, slots=True)
class Mount:
    host_path: Path
    kernel_path: Path
    permission: MountPermission

    def __str__(self):
        return f'{self.host_path}:{self.kernel_path}:{self.permission.value}'

    @classmethod
    def from_str(cls, s):
        hp, kp, perm = s.split(':')
        hp = Path(hp)
        kp = Path(kp)
        perm = MountPermission(perm)
        return cls(hp, kp, perm)


@attr.s(auto_attribs=True, slots=True)
class KernelResourceSpec:
    shares: Mapping[str, Container[Share]]
    memory_limit: int = None
    numa_node: int = None
    cpu_set: Container[int] = attr.Factory(set)
    mounts: Sequence[str] = attr.Factory(list)
    scratch_disk_size: int = None

    reserved_share_types = frozenset(['_cpu', '_mem', '_gpu', '_tpu'])

    def write_to_file(self, file: io.TextIOBase):
        '''
        Write the current resource specification into a file-like object.
        '''
        mega = lambda num_bytes: (
            f'{num_bytes // (2 ** 20)}M'
            if num_bytes > (2 ** 20) else f'{num_bytes}')
        cpu_set_str = ','.join(sorted(map(str, self.cpu_set)))
        file.write(f'NUMA_NODE={self.numa_node}\n')
        file.write(f'CPU_CORES={cpu_set_str}\n')
        file.write(f'MEMORY_LIMIT={mega(self.memory_limit)}\n')
        file.write(f'SCRATCH_SIZE={mega(self.scratch_disk_size)}\n')
        for share_type, shares in self.shares.items():
            if share_type in type(self).reserved_share_types:
                shares_str = str(shares)
            else:
                shares_str = ','.join(
                    f'{dev_id}:{dev_share}'
                    for dev_id, dev_share in shares.items())
            file.write(f'{share_type.upper()}_SHARES={shares_str}\n')
        mounts_str = ','.join(map(str, self.mounts))
        file.write(f'MOUNTS={mounts_str}\n')

    @classmethod
    def read_from_file(cls, file: io.TextIOBase):
        '''
        Read resource specification values from a file-like object.
        '''
        kvpairs = {}
        for line in file:
            key, val = line.strip().split('=', maxsplit=1)
            kvpairs[key] = val
        shares = {}
        for key, val in kvpairs.items():
            if key.endswith('_SHARES'):
                share_type = key[:-7].lower()
                if share_type in cls.reserved_share_types:
                    share_details = Decimal(val)
                else:
                    share_details = {}
                    for entry in val.split(','):
                        dev_id, dev_share = entry.split(':')
                        try:
                            dev_id = int(dev_id)
                        except ValueError:
                            pass
                        dev_share = Decimal(dev_share)
                        share_details[dev_id] = dev_share
                shares[share_type] = share_details
        mounts = [Mount.from_str(m) for m in kvpairs['MOUNTS'].split(',') if m]
        return cls(
            numa_node=int(kvpairs['NUMA_NODE']),
            cpu_set=set(map(int, kvpairs['CPU_CORES'].split(','))),
            memory_limit=readable_size_to_bytes(kvpairs['MEMORY_LIMIT']),
            scratch_disk_size=readable_size_to_bytes(kvpairs['SCRATCH_SIZE']),
            shares=shares,
            mounts=mounts,
        )

    def to_json(self):
        o = attr.asdict(self)
        o['cpu_set'] = list(sorted(o['cpu_set']))
        for dev_type, dev_shares in o['shares'].items():
            if dev_type in type(self).reserved_share_types:
                o['shares'][dev_type] = str(dev_shares)
                continue
            o['shares'][dev_type] = {
                dev_id: str(dev_share)
                for dev_id, dev_share in dev_shares.items()
            }
        o['mounts'] = list(map(str, self.mounts))
        return json.dumps(o)


class CPUAllocMap:

    def __init__(self, limit_cpus=None):
        self.limit_cpus = limit_cpus
        self.core_topo = libnuma.get_core_topology(limit_cpus)
        self.num_cores = len(libnuma.get_available_cores())
        if limit_cpus is not None:
            self.num_cores = min(self.num_cores, len(limit_cpus))
        assert sum(len(node) for node in self.core_topo) == self.num_cores
        self.num_nodes = libnuma.num_nodes()
        self.core_shares = tuple({c: 0 for c in self.core_topo[node]
                                 if limit_cpus is None or c in limit_cpus}
                                 for node in range(self.num_nodes))
        self.alloc_per_node = {n: 0 for n in range(self.num_nodes)
                               if len(self.core_shares[n]) > 0}

    def alloc(self, num_cores):
        '''
        Find a most free set of CPU cores and return a tuple of the NUMA node
        index and the set of integers indicating the found cores.
        This method guarantees that all cores are alloacted within the same
        NUMA node.
        '''
        node, current_alloc = min(
            ((n, alloc) for n, alloc in self.alloc_per_node.items()),
            key=operator.itemgetter(1))
        self.alloc_per_node[node] = current_alloc + num_cores

        shares = self.core_shares[node].copy()
        allocated_cores = set()
        for _ in range(num_cores):
            core, share = min(((core, share) for core, share in shares.items()),
                              key=operator.itemgetter(1))
            allocated_cores.add(core)
            shares[core] = sys.maxsize   # prune allocated one
            self.core_shares[node][core] += 1  # update the original share
        return node, allocated_cores

    def update(self, core_set):
        '''
        Manually add a given core set as if it is allocated by us.
        '''
        any_core = next(iter(core_set))
        node = libnuma.node_of_cpu(any_core)
        self.alloc_per_node[node] += len(core_set)
        for c in core_set:
            self.core_shares[node][c] += 1

    def free(self, core_set):
        '''
        Remove the given set of CPU cores from the allocated shares.
        It assumes that all cores are in the same NUMA node.
        '''
        any_core = next(iter(core_set))
        node = libnuma.node_of_cpu(any_core)
        self.alloc_per_node[node] -= len(core_set)
        for c in core_set:
            self.core_shares[node][c] -= 1


class AcceleratorAllocMap:

    def __init__(self,
                 devices: Collection[AbstractAcceleratorInfo],
                 limit_mask: Container[ProcessorIdType] = None):
        self._quantum = Decimal('.01')
        self.limit_mask = limit_mask
        self.devices = {dev.device_id: dev for dev in devices}
        zero = Decimal('0')
        self.device_shares = {
            p.device_id: zero for p in devices
            if limit_mask is None or p.device_id in limit_mask
        }

    def alloc(self, requested_share: Decimal, node: int = None):
        requested_share = Decimal(requested_share) \
                          .quantize(self._quantum, ROUND_DOWN)
        if node is None:
            # try finding a node, but this also may return "None"
            node = self._find_most_free_node()
        remaining_share = requested_share
        zero = Decimal('0')
        assert requested_share > zero, \
               'You cannot allocate zero share of devices.'
        allocated_shares = defaultdict(lambda: zero)
        try:
            # try filling up beginning from largest free shares
            # TODO: apply advanced scheduling and make it replacible
            while remaining_share > zero:
                p, s = self._find_largest_free_share(node)
                if s >= remaining_share:  # shortcut
                    self.device_shares[p] += remaining_share
                    allocated_shares[p] += remaining_share
                    remaining_share = zero
                elif s == 0:
                    raise RuntimeError('Cannot allocate requested shares '
                                       f'in NUMA node {node}')
                else:  # s < remaining_share
                    self.device_shares[p] += s
                    allocated_shares[p] += s
                    remaining_share -= s
        except RuntimeError:
            # revert back
            for p, s in allocated_shares.items():
                self.device_shares[p] -= s
            raise
        return node, allocated_shares

    def free(self, allocated_shares: Mapping[ProcessorIdType, Decimal]):
        for proc, share in allocated_shares.items():
            self.device_shares[proc] -= share

    def _find_most_free_node(self):
        zero = Decimal('0')
        per_node_allocs = defaultdict(lambda: zero)
        for p, s in self.device_shares.items():
            per_node_allocs[self.devices[p].numa_node] += s
        node, _ = min(
            ((n, alloc) for n, alloc in per_node_allocs.items()),
            key=operator.itemgetter(1))
        return node

    def _find_largest_free_share(self, node: int = None):
        shares = [
            (proc, self.devices[proc].max_share() - share)
            for proc, share in self.device_shares.items()
            if node is None or self.devices[proc].numa_node == node
        ]
        largest_proc_share = max(shares, key=operator.itemgetter(1))
        return largest_proc_share


def bitmask2set(mask):
    bpos = 0
    bset = []
    while mask > 0:
        if (mask & 1) == 1:
            bset.append(bpos)
        mask = (mask >> 1)
        bpos += 1
    return frozenset(bset)


async def detect_slots(etcd, limit_cpus=None, limit_gpus=None):
    '''
    Detect available resource slots of the system.
    '''

    mem_bytes = psutil.virtual_memory().total
    num_cores = len(libnuma.get_available_cores())
    if limit_cpus is not None:
        num_cores = min(num_cores, len(limit_cpus))
    slots = {
        'mem': mem_bytes >> 20,  # MiB
        'cpu': num_cores,        # core count
    }
    entry_prefix = 'backendai_accelerator_v10'
    for entrypoint in pkg_resources.iter_entry_points(entry_prefix):
        log.info(f'loading accelerator plugin: {entrypoint.module_name}')
        plugin = entrypoint.load()
        await plugin.init(etcd)
    for accel_type, accel in accelerator_types.items():
        total_share = sum(
            dev.max_share() for dev in accel.list_devices()
            if limit_gpus is None or dev.device_id not in limit_gpus)
        slots[accel.slot_key] = str(total_share)
    log.info(f'Resource slots: {slots!r}')
    return slots



from abc import abstractmethod, ABCMeta
import decimal
from pathlib import Path
from typing import (
    Any, Collection, Container, Hashable,
    Mapping, Sequence, TypeVar, Optional,
)

import aiodocker
import attr

ProcessorIdType = TypeVar('ProcessorIdType', int, str, Hashable)


accelerator_types = {}


@attr.s(auto_attribs=True)
class AbstractAcceleratorInfo(metaclass=ABCMeta):
    device_id: ProcessorIdType
    hw_location: str            # either PCI bus ID or arbitrary string
    numa_node: Optional[int]    # NUMA node ID (None if not applicable)
    memory_size: int            # bytes of available per-accelerator memory
    processing_units: int       # number of processing units (e.g., cores, SMP)

    @abstractmethod
    def is_fractional(self) -> bool:
        return False

    @abstractmethod
    def max_share(self) -> decimal.Decimal:
        '''
        Calculates the amount of available shares for this accelerator.
        '''
        return decimal.Decimal('0')

    @abstractmethod
    def share_to_spec(self, share: decimal.Decimal) -> (int, int):
        '''
        Calculates the amount of memory and processing units
        corresponding to the given share.
        '''
        return 0, 0

    @abstractmethod
    def spec_to_share(self, requested_memory: int,
                      requested_proc_units: int) -> decimal.Decimal:
        '''
        Calculates the minimum required share to guarantee the given
        requested resource specification (memory in bytes and the number
        of processing units).
        '''
        return decimal.Decimal('0')


class AbstractAccelerator(metaclass=ABCMeta):

    slot_key = 'accelerator'

    @classmethod
    @abstractmethod
    def list_devices(cls) -> Collection[AbstractAcceleratorInfo]:
        '''
        Return a collection of processors.
        '''
        return []

    @classmethod
    @abstractmethod
    def get_hooks(cls, distro: str, arch: str) -> Sequence[Path]:
        '''
        Return the library hook paths used by the plugin (optional).

        :param str distro: The target Linux distribution such as "ubuntu16.04" or
                           "alpine3.8"
        :param str arch: The target CPU architecture such as "amd64"
        '''
        return []

    @abstractmethod
    async def generate_docker_args(
            cls,
            docker: 'aiodocker.docker.Docker',  # noqa
            limit_gpus: Container[ProcessorIdType] = None) \
            -> Mapping[str, Any]:
        return {}



try:
    import builtins
except ImportError:
    builtins = __builtins__
import socket
import sys

input_host = '127.0.0.1'
input_port = 65000


def _input(prompt=''):
    sys.stdout.write(prompt)
    sys.stdout.flush()
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        try:
            sock.connect((input_host, input_port))
            userdata = sock.recv(1024)
        except ConnectionRefusedError:
            userdata = b'<user-input-unavailable>'
    return userdata.decode()


builtins._input = input
builtins.input = _input



'''
Displays Agg images in the browser, wrapping them as Sorna media responses.
'''

import base64
import io

# Base imports from matplotlib.backends.backend_template
import matplotlib  # noqa
from matplotlib._pylab_helpers import Gcf
from matplotlib.backend_bases import FigureManagerBase
from matplotlib.figure import Figure

# The real renderer that generates image data
_backend = 'svg'  # noqa
from matplotlib.backends import backend_agg
from matplotlib.backends import backend_svg

# My own imports
from six.moves import builtins

from ..types import MediaRecord


def draw_if_interactive():
    pass


def show():
    for manager in Gcf.get_all_fig_managers():
        manager.show()


def new_figure_manager(num, *args, **kwargs):
    FigureClass = kwargs.pop('FigureClass', Figure)
    thisFig = FigureClass(*args, **kwargs)
    return new_figure_manager_given_figure(num, thisFig)


def new_figure_manager_given_figure(num, figure):
    canvas = FigureCanvas(figure)
    manager = FigureManager(canvas, num)
    return manager


class FigureCanvasSornaAgg(backend_agg.FigureCanvasAgg):
    supports_blit = False

    def __init__(self, *args, **kwargs):
        super(FigureCanvasSornaAgg, self).__init__(*args, **kwargs)
        self._is_old = True
        self._dpi_ratio = 1

    def draw(self):
        renderer = self.get_renderer(cleared=True)
        self._is_old = True
        backend_agg.RendererAgg.lock.acquire()
        try:
            self.figure.draw(renderer)
        finally:
            backend_agg.RendererAgg.lock.release()

    def get_default_filetype(self):
        return 'png'


class FigureCanvasSornaSVG(backend_svg.FigureCanvasSVG):

    def __init__(self, *args, **kwargs):
        super(FigureCanvasSornaSVG, self).__init__(*args, **kwargs)


class FigureManagerSorna(FigureManagerBase):

    def __init__(self, *args, **kwargs):
        super(FigureManagerSorna, self).__init__(*args, **kwargs)

    def show(self):
        if _backend == 'agg':
            with io.BytesIO() as buf:
                self.canvas.print_png(buf)
                raw = buf.getvalue()
            enc = base64.b64encode(raw)
            builtins._sorna_emit(MediaRecord(
                'image/png',
                'data:image/png;base64,' + enc.decode('ascii')
            ))
        elif _backend == 'svg':
            with io.BytesIO() as buf:
                self.canvas.print_svg(buf)
                raw = buf.getvalue()
            builtins._sorna_emit(MediaRecord(
                'image/svg+xml',
                raw.decode('utf8'),
            ))
        else:
            raise RuntimeError('Unsupported sorna matplotlib backend.')


if _backend == 'svg':
    FigureCanvas = FigureCanvasSornaSVG
elif _backend == 'agg':
    FigureCanvas = FigureCanvasSornaAgg
else:
    raise ImportError('_backend has wrong value!')

FigureManager = FigureManagerSorna
